;;;-*-Mode: LISP; Package: CCL -*-
;;;
;;;   Copyright (C) 2001 Clozure Associates
;;;   This file is part of Opensourced MCL.
;;;
;;;   Opensourced MCL is free software; you can redistribute it and/or
;;;   modify it under the terms of the GNU Lesser General Public
;;;   License as published by the Free Software Foundation; either
;;;   version 2.1 of the License, or (at your option) any later version.
;;;
;;;   Opensourced MCL is distributed in the hope that it will be useful,
;;;   but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;;   Lesser General Public License for more details.
;;;
;;;   You should have received a copy of the GNU Lesser General Public
;;;   License along with this library; if not, write to the Free Software
;;;   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;;;


;;; GDBM is licensed under the GPL, so we try to make its use
;;; slightly optional.
;;; A better way around this issue would be to use something
;;; other than GDBM ...

(defvar *use-gdbm* t)

(defloadvar *gdbm-lib*
    (and *use-gdbm*
	 (ignore-errors (open-shared-library "libgdbm.so"))))


(defstruct ffi-type
  (ordinal nil)
  (defined nil)
  (string)
  (name)                                ; a keyword, uppercased or NIL
)

(defmethod print-object ((x ffi-type) out)
  (print-unreadable-object (x out :type t :identity t)
    (format out "~a" (ffi-type-string x))))

(defvar *ffi-prefix* "")


(defstruct (ffi-union (:include ffi-type))
  fields
  (anon-global-id (unless name (concatenate 'string *ffi-prefix* "-" string))))

(defstruct (ffi-struct (:include ffi-type))
  fields
  (anon-global-id (unless name (concatenate 'string *ffi-prefix* "-" string))))

(defstruct (ffi-typedef (:include ffi-type))
  (type))

(defun ffi-struct-reference (s)
  (or (ffi-struct-name s) (ffi-struct-anon-global-id s)))

(defun ffi-union-reference (u)
  (or (ffi-union-name u) (ffi-union-anon-global-id u)))

(defstruct (ffi-function (:include ffi-type))
  arglist
  return-value)

(def-foreign-type :datum 
 (:struct nil  
  (:dptr (* (:unsigned 8) ) )
  (:dsize (:signed 32) )) )

(def-foreign-type
  :gdbm_file
    (* (:struct nil (:dummy (:array (:signed 32)  10) )) ) )

    
(def-foreign-type nil
  (:struct dbm-constant
   (:class (:unsigned 32))
   (:value
    (:union nil
      (:s32 (:signed 32))
      (:u32 (:unsigned 32))
      (:single-float :float)
      (:double-float :double)))))

(defconstant db-string-constant 0)
(defconstant db-read-string-constant 1)
(defconstant db-s32-constant 2)
(defconstant db-u32-constant 3)
(defconstant db-float-constant 4)
(defconstant db-double-constant 5)
(defconstant db-char-constant 6)

(defparameter *arg-spec-encoding*
  '((#\Space . :void)
    (#\a . :address)
    (#\F . :signed-fullword)
    (#\f . :unsigned-fullword)
    (#\H . :signed-halfword)
    (#\h . :unsigned-halfword)
    (#\B . :signed-byte)
    (#\b . :unsigned-byte)
    (#\s . :single-float)
    (#\d . :double-float)))

(defun encode-arguments (args result string)
  (let* ((i 1))
    (flet ((encoded-arg (thing)
             (if thing
               (or (car (rassoc thing *arg-spec-encoding*))
                   (return-from encode-arguments nil))
               #\Space)))
      (setf (schar string 0) (encoded-arg result))
      (dolist (arg args t)
        (setf (schar string i) (encoded-arg arg)
              i (1+ i))))))

(defun decode-arguments (string)
  (let* ((result (cdr (assoc (schar string 0) *arg-spec-encoding*)))
         (args ()))
    (do* ((i 1 (1+ i)))
         ((= i (length string)) (values (nreverse args) result))
      (push (cdr (assoc (schar string i) *arg-spec-encoding*))
            args))))

;;; encoded external function looks like:
;;; byte min-args
;;; byte name-length
;;; name-length bytes of name
;;; result+arg specs

(defun extract-db-function (datum)
  (let* ((val nil)
         (dsize (pref datum :datum.dsize)))
    (with-macptrs ((dptr))
      (%setf-macptr dptr (pref datum :datum.dptr))
      (unless (%null-ptr-p dptr)
        (unwind-protect
         (let* ((min-args (%get-byte dptr))
                (name-len (%get-byte dptr 1))
                (external-name (%str-from-ptr (%inc-ptr dptr 2) name-len))
                (encoding-len (- dsize (+ 2 name-len)))
                (encoding (make-string encoding-len)))
           (declare (dynamic-extent encoding))
           (%copy-ptr-to-ivector dptr (+ 2 name-len) encoding 0 encoding-len)
           (multiple-value-bind (args result)
               (decode-arguments encoding)
             (setq val (make-external-function-definition
                        :entry-name external-name
                        :arg-specs args
                        :result-spec result
                        :min-args min-args))))
          (free dptr))))
    val))

(defun db-lookup-function (gdbm-file name)
  (rlet ((value :datum)
         (key datum))
    (with-cstrs ((keyname (string name)))
      (setf (pref key :datum.dptr) keyname
            (pref key :datum.dsize) (length (string name))
            (pref value :datum.dptr) (%null-ptr)
            (pref value :datum.dsize) 0)
      (#_gdbm_fetch value gdbm-file key)
      (extract-db-function value))))


(defun save-db-function (gdbm-file fname def)
  (let* ((external-name (efd-entry-name def))
         (args (efd-arg-specs def))
         (result (efd-result-spec def))
         (min-args (efd-min-args def))
         (namelen (length external-name))
         (enclen (1+ (length args)))
         (buflen (+ 2 namelen enclen))
         (encoding (make-string enclen)))
    (declare (dynamic-extent encoding))
    (when (encode-arguments args result encoding)
      (%stack-block ((buf buflen))
        (setf (%get-byte buf) min-args
              (%get-byte buf 1) namelen)
        (%copy-ivector-to-ptr external-name 0 buf 2 namelen)
        (%copy-ivector-to-ptr encoding 0 buf (+ 2 namelen) enclen)
        (rlet ((contents :datum)
               (key :datum))
          (setf (pref contents :datum.dptr) buf
                (pref contents :datum.dsize) buflen)
          (with-cstrs ((keyname (string fname)))
            (setf (pref key :datum.dptr) keyname
                  (pref key :datum.dsize) (length (string fname)))
	    (#_gdbm_store gdbm-file key contents #$GDBM_REPLACE)))))))
        
         
(defun extract-db-constant-value (datum)
  (let* ((val nil)
         (dsize (pref datum :datum.dsize)))
    (with-macptrs ((dptr))
      (%setf-macptr dptr (pref datum :datum.dptr))
      (unless (%null-ptr-p dptr)
        (unwind-protect
             (let* ((class (pref dptr :dbm-constant.class)))
               (setq val
                     (ecase class
                  ((#.db-string-constant #.db-read-string-constant)
                   (let* ((str (%str-from-ptr (%inc-ptr dptr 4) (- dsize 4))))
                     (if (eql class db-read-string-constant)
                       (read-from-string str)
                       str)))
                  (#.db-s32-constant (pref dptr :dbm-constant.value.s32))
                  (#.db-u32-constant (pref dptr :dbm-constant.value.u32))
                  (#.db-float-constant (pref dptr :dbm-constant.value.single-float))
                  (#.db-double-constant (pref dptr :dbm-constant.value.double-float))
                  (#.db-char-constant (code-char (pref dptr :dbm-constant.value.u32))))))
          (free dptr))))
    val))

(defun db-lookup-constant (gdbm-file name)
  (rlet ((value :datum)
         (key datum))
    (with-cstrs ((keyname (string name)))
      (setf (pref key :datum.dptr) keyname
            (pref key :datum.dsize) (length (string name))
            (pref value :datum.dptr) (%null-ptr)
            (pref value :datum.dsize) 0)
      (#_gdbm_fetch value gdbm-file key)
      (extract-db-constant-value value))))
    
(defun %gdbm-open (namestring flags mode &optional (block-size 0))
  (with-cstrs ((dbname namestring))
    (let* ((result
            (#_gdbm_open dbname block-size flags mode (%null-ptr))))
      (unless (%null-ptr-p result)
        result))))

(defun create-gdbm-file (pathname)
  (or (%gdbm-open (namestring (translate-logical-pathname pathname))  #$GDBM_NEWDB #o666)
      (error "Couldn't create gdbm file ~s" pathname)))

(defun open-input-gdbm-file (pathname)
  (or
   (%gdbm-open (namestring (translate-logical-pathname pathname)) #$GDBM_READER  #o666)
   (error "Couldn't open database file ~s" pathname )))

(defun %gdbm-close (gdbm-file)
  (#_gdbm_close gdbm-file))      



(defun db-define-string-constant (gdbm-file name val &optional (class db-string-constant))
  (let* ((dsize (+ 4 (length val))))
    (%stack-block ((valbuf dsize))
      (%copy-ivector-to-ptr val 0 valbuf 4 (length val))
      (setf (%get-long valbuf) class)
      (rlet ((content :datum)
             (key :datum))
        (setf (pref content :datum.dsize) dsize
              (pref content :datum.dptr) valbuf)
        (with-cstrs ((keyname (string name)))
          (setf (pref key :datum.dsize) (length (string name))
                (pref key :datum.dptr) keyname)
          (#_gdbm_store gdbm-file key content #$GDBM_REPLACE))))))
      
(defun db-define-constant (gdbm-file name val)
  (typecase val
    (string (db-define-string-constant gdbm-file name val))
    ((or (unsigned-byte 32)
         (signed-byte 32)
         short-float
         double-float
         character)
     (rlet ((constant :dbm-constant)
            (content :datum)
            (key :datum))
       (etypecase val
         ((signed-byte 32)
          (setf (pref constant :dbm-constant.value.s32) val)
          (setf (pref constant :dbm-constant.class) db-s32-constant))
         ((unsigned-byte 32)
          (setf (pref constant :dbm-constant.value.u32) val)
          (setf (pref constant :dbm-constant.class) db-u32-constant))
         (short-float
          (setf (pref constant :dbm-constant.value.single-float) val)
          (setf (pref constant :dbm-constant.class) db-float-constant))
         (double-float
          (setf (pref constant :dbm-constant.value.double-float) val)
          (setf (pref constant :dbm-constant.class) db-double-constant))
         (character
          (setf (pref constant :dbm-constant.value.u32) (char-code val))
          (setf (pref constant :dbm-constant.class) db-char-constant)))
       (setf (pref content :datum.dptr) constant
             (pref content :datum.dsize) (record-length :dbm-constant))
       (with-cstrs ((keyname (string name)))
         (setf (pref key :datum.dptr) keyname
               (pref key :datum.dsize) (length (string name)))
         (#_gdbm_store gdbm-file key content #$GDBM_REPLACE))))
    (t (db-define-string-constant gdbm-file name (format nil "~a" val) db-read-string-constant))))

(defmacro with-new-gdbm-file ((var pathname) &body body)
  (let* ((gdbm (gensym)))
    `(let* (,gdbm)
      (unwind-protect
           (let* ((,var (setq ,gdbm (create-gdbm-file ,pathname))))
             ,@body)
        (when ,gdbm (%gdbm-close ,gdbm))))))

(defvar *os-package* (find-package "OS"))


(defloadvar *gdbm-constants* nil)
(defun gdbm-constants ()
  (or *gdbm-constants*
      (setq *gdbm-constants* (open-input-gdbm-file "ccl:headers;constants.gdbm"))))

(defloadvar *gdbm-types* nil)
(defun gdbm-types ()
  (or *gdbm-types*
      (setq *gdbm-types* (open-input-gdbm-file "ccl:headers;types.gdbm"))))

(defloadvar *gdbm-records* nil)
(defun gdbm-records ()
  (or *gdbm-records*
      (setq *gdbm-records* (open-input-gdbm-file "ccl:headers;records.gdbm"))))

(defloadvar *gdbm-functions* nil)
(defun gdbm-functions ()
  (or *gdbm-functions*
      (setq *gdbm-functions* (open-input-gdbm-file "ccl:headers;functions.gdbm"))))

(defun load-os-constant (sym &optional reader-stream)
  (declare (ignore reader-stream))
  (let* ((val (or (db-lookup-constant (gdbm-constants) sym)
                  (error "Constant not found: ~s" sym))))
    (let* ((*record-source-file* nil))
      (%defconstant sym val)
      val)))

(defun load-external-function (sym &optional reader-stream)
  (declare (ignore reader-stream))
  (let* ((def (or (db-lookup-function (gdbm-functions) sym)
                  (error "Foreign function not found: ~s" sym))))
    (setf (gethash sym *external-function-definitions*) def)
    (setf (macro-function sym) #'%external-call-expander)
    sym))
  
(set-dispatch-macro-character 
 #\# #\$
 (qlfun |#$-reader| (stream char arg)
   (declare (special *os-package*))
   (declare (ignore char))
   (let* ((sym (let ((*package* *os-package*)) (read stream t nil t))))
     (unless *read-suppress*
       (unless arg (setq arg 0))
       (ecase arg
         (0
          (unless (boundp sym)
            (load-os-constant sym stream)))
         (1 (makunbound sym) (load-os-constant sym stream)))
       sym))))

(set-dispatch-macro-character #\# #\_
  (qlfun |#_-reader| (stream char arg)
    (when arg (unless *read-suppress* (%err-disp $xrdnoarg char)))
    (let* ((sym (let ((*package* *os-package*)) (read stream t nil t))))
      (unless *read-suppress*
        (unless (and sym (symbolp sym)) (report-bad-arg sym 'symbol))
        (load-external-function sym stream)))))



(eval-when (:compile-toplevel :execute)
  (defconstant encoded-type-void 0)
  (defconstant encoded-type-signed-32 1)
  (defconstant encoded-type-unsigned-32 2)
  (defconstant encoded-type-signed-8 3)
  (defconstant encoded-type-unsigned-8 4)
  (defconstant encoded-type-signed-16 5)
  (defconstant encoded-type-unsigned-16 6)
  (defconstant encoded-type-signed-n 7) ;N
  (defconstant encoded-type-unsigned-n 8) ;N
  (defconstant encoded-type-single-float 9)
  (defconstant encoded-type-double-float 10)
  (defconstant encoded-type-pointer 11) ; <type>
  (defconstant encoded-type-array 12) ; <size> <type>
  (defconstant encoded-type-named-struct-ref 13); <tag>
  (defconstant encoded-type-named-union-ref 14) ;<tag>
  (defconstant encoded-type-named-type-ref 15) ; <name>
  (defconstant encoded-type-anon-struct-ref 16) ; <tag>
  (defconstant encoded-type-anon-union-ref 17) ; <tag>
  )


(defun encode-name (name)
  (if (null name)
    '(0)
    (let* ((string (string name))
           (length (length string)))
      (cons length (map 'list #'char-code string)))))

(defun encode-ffi-field (field)
  `(,@(encode-name (car field)) ,@(encode-ffi-type (cadr field))))

(defun encode-ffi-field-list (fields)
  (let* ((len (length fields)))
    (labels ((encode-fields (fields)
               (if fields
                 `(,@(encode-ffi-field (car fields)) ,@(encode-fields (cdr fields))))))
      `(,len ,@(encode-fields fields)))))

(defun encode-ffi-union (u)
  (let* ((name (ffi-union-name u)))
    (if name
      `(,encoded-type-named-union-ref 
        ,@(encode-name name)
        ,@(encode-ffi-field-list (ffi-union-fields u)))
      `(,encoded-type-anon-union-ref
        ,@(encode-ffi-field-list (ffi-union-fields u))))))

(defun encode-ffi-struct (s)
  (let* ((name (ffi-struct-name s)))
    (if name
      `(,encoded-type-named-struct-ref 
        ,@(encode-name (ffi-struct-name s))
        ,@(encode-ffi-field-list (ffi-struct-fields s)))
      `(,encoded-type-anon-struct-ref
        ,@(encode-ffi-field-list (ffi-struct-fields s))))))

(defun encode-u32 (val)
  `(,(ldb (byte 8 24) val)
    ,(ldb (byte 8 16) val)
    ,(ldb (byte 8 8) val)
    ,(ldb (byte 8 0) val)))


(defun encode-ffi-type (spec)
  (case (car spec)
    (:primitive
     (let ((primtype (cadr spec)))
       (if (atom primtype)
         (case primtype
           (:float `(,encoded-type-single-float))
           (:double `(,encoded-type-double-float))
           (:void `(,encoded-type-void))
           (:signed `(,encoded-type-signed-32))
           (:unsigned `(,encoded-type-unsigned-32))
           ((:long-long :long-double :complex-int
                        :complex-float :complex-double :complex-long-double)
            (encode-ffi-type `(:struct ,primtype))))
         (ecase (car primtype)
           (* `(,encoded-type-pointer ,@(encode-ffi-type
                                           (if (eq (cadr primtype) t)
                                             `(:primitive :void)
                                             (cadr primtype)))))
           (:signed
            (case (cadr primtype)
              (32 `(,encoded-type-signed-32))
              (16 `(,encoded-type-signed-16))
              (8 `(,encoded-type-signed-8))
              (t `(,encoded-type-signed-n ,(cadr primtype)))))
           (:unsigned
            (case (cadr primtype)
              (32 `(,encoded-type-unsigned-32))
              (16 `(,encoded-type-unsigned-16))
              (8 `(,encoded-type-unsigned-8))
              (t `(,encoded-type-unsigned-n ,(cadr primtype)))))))))
     (:struct
      (let* ((s (cadr spec))
             (name (ffi-struct-name s)))
      `(,(if name
             encoded-type-named-struct-ref
             encoded-type-anon-struct-ref)
        ,@(encode-name (ffi-struct-reference s)))))
     (:union
      (let* ((u (cadr spec))
             (name (ffi-union-name u)))
      `(,(if name
             encoded-type-named-union-ref
             encoded-type-anon-union-ref)
        ,@(encode-name (ffi-union-reference u)))))
     (:typedef
      `(,encoded-type-named-type-ref ,@(encode-name (ffi-typedef-name (cadr spec)))))
     (:pointer
      `(,encoded-type-pointer ,@(encode-ffi-type
                                   (if (eq (cadr spec) t)
                                     '(:primitive :void)
                                     (cadr spec)))))
     (:array
      `(,encoded-type-array ,@(encode-u32 (cadr spec)) ,@(encode-ffi-type (caddr spec))))
     (t
      (break "Type spec = ~s" spec))))

(defun encode-ffi-arg-type (spec)
  (case (car spec)
    (:primitive
     (let ((primtype (cadr spec)))
       (if (atom primtype)
         (case primtype
           (:float `(#\s))
           (:double `(#\d))
           (:void `(#\Space))
           (:signed `(#\F))
           (:unsigned `(f))
           ((:long-long :long-double :complex-int
                        :complex-float :complex-double :complex-long-double)            
            #|(encode-ffi-arg-type `(:struct ,primtype))|#
            `(#\?)))
         (ecase (car primtype)
           (* `(#\a))
           (:signed
            (let* ((nbits (cadr primtype)))
              (if (<= nbits 8)
                '(#\B)
                (if (<= nbits 16)
                  '(#\H)
                  (if (<= nbits 32)
                    '(#\F)
                    '(#\?))))))
           (:unsigned
            (let* ((nbits (cadr primtype)))
              (if (<= nbits 8)
                '(#\b)
                (if (<= nbits 16)
                  '(#\h)
                  (if (<= nbits 32)
                    '(#\f)              
                    '(#\?))))))))))
    ((:struct :union) `(#\a))
    (:typedef
      (encode-ffi-arg-type (ffi-typedef-type (cadr spec))))
    (:pointer
      `(#\a))
    (:array
      `(#\?))))

(defun encode-ffi-arg-list (args)
  (if args
    `(,@(encode-ffi-arg-type (car args)) ,@(encode-ffi-arg-list (cdr args)))))

(defun encode-ffi-function (f)
  (let* ((args (ffi-function-arglist f))
         (min-args (length args))
         (result (ffi-function-return-value f)))
    `(,min-args
      ,@(encode-name (ffi-function-string f))
      ,@(encode-ffi-arg-type result)
      ,@(encode-ffi-arg-list args))))
    
(defun save-byte-list (ptr l)
  (do* ((l l (cdr l))
        (i 0 (1+ i)))
       ((null l))
    (let* ((b (car l)))
      (if (typep b 'character)
        (setq b (char-code b)))
      (setf (%get-unsigned-byte ptr i) b))))

(defun gdbm-write-byte-list (gdbm-file keyname bytes)
  (let* ((len (length bytes)))
    (%stack-block ((p len))
      (save-byte-list p bytes)
      (rlet ((contents :datum)
             (key :datum))
        (with-cstrs ((keystring (string keyname)))
          (setf (pref contents :datum.dptr) p
                (pref contents :datum.dsize) len
                (pref key :datum.dptr) keystring
                (pref key :datum.dsize) (length (string keyname)))
          (#_gdbm_store gdbm-file key contents #$GDBM_REPLACE))))))

(defun save-ffi-function (gdbm-file fun)
  (let* ((encoding (encode-ffi-function fun)))
    (gdbm-write-byte-list gdbm-file
                          (string-upcase (ffi-function-string fun))
                          encoding)))

(defun save-ffi-typedef (gdbm-file def)
  (gdbm-write-byte-list gdbm-file
                       (ffi-typedef-name def)
                       (encode-ffi-type (ffi-typedef-type def))))

(defun save-ffi-struct (gdbm-file s)
  (gdbm-write-byte-list gdbm-file (ffi-struct-reference s) (encode-ffi-struct s)))

(defun save-ffi-union (gdbm-file u)
  (gdbm-write-byte-list gdbm-file (ffi-union-reference u) (encode-ffi-union u)))

(defun %decode-name (buf p)
  (declare (type macptr buf) (fixnum p))
  (let* ((n (%get-unsigned-byte buf p)))
    (declare (fixnum n))
    (if (zerop n)
      (values nil 1)
      (let* ((pname (make-string n)))
        (%copy-ptr-to-ivector buf (1+ p) pname 0 n)
        (values (intern pname *keyword-package*)
                (+ p (1+ n)))))))

(defun %decode-u32 (buf p)
  (declare (fixnum p) (type macptr buf))
  (values (dpb
           (%get-unsigned-byte buf p)
           (byte 8 24)
           (dpb
            (%get-unsigned-byte buf (+ p 1))
            (byte 8 16)
            (dpb
             (%get-unsigned-byte buf (+ p 2))
             (byte 8 8)
             (%get-unsigned-byte buf (+ p 3)))))
          (+ p 4)))
  
;; Should return a FOREIGN-TYPE structure.
(defun %decode-type (buf p)
  (declare (type macptr buf) (fixnum p))
  (let* ((q (1+ p)))
    (ecase (%get-unsigned-byte buf p)
      (#.encoded-type-void (values (parse-foreign-type :void) q))
      (#.encoded-type-signed-32 (values (svref *signed-integer-types* 32) q))
      (#.encoded-type-unsigned-32 (values (svref *unsigned-integer-types* 32) q))
      (#.encoded-type-signed-8 (values (svref *signed-integer-types* 8) q))
      (#.encoded-type-unsigned-8 (values (svref *unsigned-integer-types* 8) q))
      (#.encoded-type-signed-16 (values (svref *signed-integer-types* 16) q))
      (#.encoded-type-unsigned-16 (values (svref *unsigned-integer-types* 16) q))
      (#.encoded-type-signed-n (values (make-foreign-integer-type
                                          :signed t
                                          :bits (%get-unsigned-byte buf q))
                                         (1+ q)))
      (#.encoded-type-unsigned-n (values (make-foreign-integer-type
                                            :signed nil
                                            :bits (%get-unsigned-byte buf q))
                                           (1+ q)))
      (#.encoded-type-single-float (values (parse-foreign-type :float) q))
      (#.encoded-type-double-float (values (parse-foreign-type :double) q))
      (#.encoded-type-pointer (multiple-value-bind (target qq)
                                    (if (eql (%get-unsigned-byte buf q)
                                             encoded-type-void)
                                      (values nil (1+ q))
                                      (%decode-type buf q))
                                  (values (make-foreign-pointer-type :to target)
                                          qq)))
      (#.encoded-type-array
       (multiple-value-bind (size qq) (%decode-u32 buf q)
         (multiple-value-bind (target qqq) (%decode-type buf qq)
           (let* ((type-alignment (foreign-type-alignment target))
                  (type-bits (foreign-type-bits target)))
             (values (make-foreign-array-type
                      :element-type target
                      :dimensions (list size)
                      :alignment type-alignment
                      :bits (if type-bits
                              (* (align-offset type-bits type-alignment) size)))
                     qqq)))))
      (#.encoded-type-named-type-ref
       (multiple-value-bind (name qq) (%decode-name buf q)
         (values (%parse-foreign-type name) qq)))
      (#.encoded-type-named-struct-ref
       (multiple-value-bind (name qq) (%decode-name buf q)
         (values (or (info-foreign-type-struct name)
                     (setf (info-foreign-type-struct name)
                           (make-foreign-record-type :kind :struct
                                                     :name name)))
                 qq)))
      (#.encoded-type-named-union-ref
       (multiple-value-bind (name qq) (%decode-name buf q)
         (values (or (info-foreign-type-union name)
                     (setf (info-foreign-type-union name)
                           (make-foreign-record-type :kind :union
                                                     :name name)))
                 qq)))
      ((#.encoded-type-anon-struct-ref #.encoded-type-anon-union-ref)
       (multiple-value-bind (tag qq) (%decode-name buf q)
         (values (load-record tag) qq))))))


(defun extract-db-type (datum)
  (let* ((type nil))
    (with-macptrs ((dptr))
      (%setf-macptr dptr (pref datum :datum.dptr))
      (unless (%null-ptr-p dptr)
        (unwind-protect
             (setq type (%decode-type dptr 0))
          (free dptr))))
    type))

(defun %load-foreign-type (gdbm-file name)
  (with-cstrs ((string (string name)))
    (rlet ((contents :datum)
           (key :datum))
      (setf (pref key :datum.dsize) (length (string name))
            (pref key :datum.dptr) string
            (pref contents :datum.dptr) (%null-ptr)
            (pref contents :datum.dsize) 0)
      (#_gdbm_fetch contents gdbm-file key)
      (let* ((type (extract-db-type contents)))
        (if type
          (%def-foreign-type name type))))))

(defun load-foreign-type (name)
  (%load-foreign-type (gdbm-types) name))

(defun %decode-field (buf p)
  (declare (type macptr buf) (fixnum p))
  (multiple-value-bind (name q) (%decode-name buf p)
    (multiple-value-bind (type r) (%decode-type buf q)
      (values (make-foreign-record-field :type type :name name) r))))

(defun %decode-field-list (buf p)
  (declare (type macptr buf) (fixnum p))
  (let* ((n (%get-unsigned-byte buf p))
         (fields nil))
    (incf p)
    (dotimes (i n (values (nreverse fields) p))
      (multiple-value-bind (field q) (%decode-field buf p)
        (push field fields)
        (setq p q)))))

(defun %determine-record-attributes (rtype parsed-fields)
  (let* ((total-bits 0)
         (overall-alignment 1)
         (kind (foreign-record-type-kind rtype)))
    (dolist (field parsed-fields)
      (let* ((field-type (foreign-record-field-type field))
             (bits (ensure-foreign-type-bits field-type))
             (alignment (foreign-type-alignment field-type)))
        (unless bits
          (error "Unknown size: ~S"
                 (unparse-foreign-type field-type)))
        (unless alignment
          (error "Unknown alignment: ~S"
                 (unparse-foreign-type field-type)))
        (setq overall-alignment (max overall-alignment alignment))
        (ecase kind
          (:struct (let* ((offset (align-offset total-bits alignment)))
                     (setf (foreign-record-field-offset field) offset)
                     (setq total-bits (+ offset bits))))
          (:union (setq total-bits (max total-bits bits))))))
    (setf (foreign-record-type-fields rtype) parsed-fields
          (foreign-record-type-alignment rtype) overall-alignment
          (foreign-record-type-bits rtype) (align-offset total-bits overall-alignment))
    rtype))

(defun %decode-record-type (buf p)
  (declare (type macptr buf) (fixnum p))
  (let* ((rcode (%get-unsigned-byte buf p)))
    (multiple-value-bind (name q)
        (case rcode
          ((#.encoded-type-anon-struct-ref #.encoded-type-anon-union-ref)
           (values nil (1+ p)))
          (t
           (%decode-name buf (1+ p))))
      (%determine-record-attributes
       (if name
         (if (eql rcode encoded-type-named-struct-ref)
           (or (info-foreign-type-struct name)
               (setf (info-foreign-type-struct name)
                     (make-foreign-record-type :kind :struct :name name)))
           (or (info-foreign-type-union name)
               (setf (info-foreign-type-union name)
                     (make-foreign-record-type :kind :union :name name))))
         (make-foreign-record-type
          :kind (if (eql rcode encoded-type-anon-struct-ref)
                  :struct
                  :union)
          :name name))
       (%decode-field-list buf q)))))

(defun extract-db-record (datum)
  (let* ((type nil))
    (with-macptrs ((dptr))
      (%setf-macptr dptr (pref datum :datum.dptr))
      (unless (%null-ptr-p dptr)
        (unwind-protect
             (setq type (%decode-record-type dptr 0))
          (free dptr))))
    type))

(defun %load-foreign-record (gdbm-file name)
  (with-cstrs ((string (string name)))
    (rlet ((contents :datum)
           (key :datum))
      (setf (pref key :datum.dsize) (length (string name))
            (pref key :datum.dptr) string
            (pref contents :datum.dptr) (%null-ptr)
            (pref contents :datum.dsize) 0)
      (#_gdbm_fetch contents gdbm-file key)
      (extract-db-record contents))))

(defun load-record (name)
  (%load-foreign-record (gdbm-records) name))


