;;;-*-Mode: LISP; Package: CCL -*-
;;;
;;;   Copyright (C) 2001 Clozure Associates
;;;   This file is part of Opensourced MCL.
;;;
;;;   Opensourced MCL is free software; you can redistribute it and/or
;;;   modify it under the terms of the GNU Lesser General Public
;;;   License as published by the Free Software Foundation; either
;;;   version 2.1 of the License, or (at your option) any later version.
;;;
;;;   Opensourced MCL is distributed in the hope that it will be useful,
;;;   but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;;   Lesser General Public License for more details.
;;;
;;;   You should have received a copy of the GNU Lesser General Public
;;;   License along with this library; if not, write to the Free Software
;;;   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;;;

(in-package "CCL")

(defstruct linux-syscall
  (idx 0 :type fixnum)
  (arg-specs () :type list)
  (result-spec nil :type symbol)
  (min-args 0 :type fixnum))

(defvar *linux-syscalls* (make-hash-table :test #'eq))

(defmacro syscall (name &rest args)
  (let* ((info (or (gethash name *linux-syscalls*)
		   (error "Unknown system call: ~s" name)))
	 (idx (linux-syscall-idx info))
	 (arg-specs (linux-syscall-arg-specs info))
	 (n-argspecs (length arg-specs))
	 (n-args (length args))
	 (min-args (linux-syscall-min-args info))
	 (result (linux-syscall-result-spec info)))
    (unless (and (>= n-args min-args) (<= n-args n-argspecs))
      (error "wrong number of args in ~s" args))
    (do* ((call ())
	  (specs arg-specs (cdr specs))
	  (args args (cdr args)))
	 ((null args)
	  `(linux-syscall ,idx ,@(nreverse (cons result call))))
      (push (car specs) call)
      (push (car args) call))))

(defmacro define-syscall (name idx (&rest arg-specs) result-spec
			       &key (min-args (length arg-specs)))
  `(progn
    (setf (gethash ',name *linux-syscalls*)
     (make-linux-syscall :idx ,idx
      :arg-specs ',arg-specs :result-spec ',result-spec :min-args ,min-args))
    ',name))

(define-syscall linux::exit 1 (:signed-fullword) :void)
(define-syscall linux::fork 2 () :signed-fullword)
(define-syscall linux::read 3 (:unsigned-fullword :address :unsigned-fullword)
		:signed-fullword)
(define-syscall linux::write 4 (:unsigned-fullword :address :unsigned-fullword)
		:signed-fullword)
(define-syscall linux::open 5 (:address :unsigned-fullword :unsigned-fullword) :signed-fullword :min-args 2)
(define-syscall linux::close 6 (:unsigned-fullword) :signed-fullword )
(define-syscall linux::waitpid 7 (:unsigned-fullword :address :signed-fullword)
		:signed-fullword)
(define-syscall linux::creat 8 (:address :unsigned-fullword) :signed-fullword )
(define-syscall linux::link 9 (:address :address) :signed-fullword)
(define-syscall linux::unlink 10 (:address) :signed-fullword )
(define-syscall linux::execve 11 (:address :address :address) :signed-fullword )
(define-syscall linux::chdir 12 (:address) :signed-fullword )
(define-syscall linux::time 13 (:address) :unsigned-fullword )
(define-syscall linux::mknod 14 (:address :unsigned-fullword :unsigned-fullword)
		:signed-fullword)
(define-syscall linux::chmod 15 (:address :unsigned-fullword) :signed-fullword)
(define-syscall linux::lchown 16 (:address :unsigned-fullword :unsigned-fullword)
		:signed-fullword)
;(define-syscall linux::oldstat 18 () )
(define-syscall linux::lseek 19 (:unsigned-fullword :signed-fullword :unsigned-fullword) :signed-fullword )
(define-syscall linux::getpid 20 () :unsigned-fullword)
(define-syscall linux::mount 21 (:address
				 :address
				 :address
				 :unsigned-fullword
				 :address) :signed-fullword )
(define-syscall linux::umount 22 (:address) :signed-fullword )

(define-syscall linux::setuid 23 (:unsigned-fullword) :signed-fullword )
(define-syscall linux::getuid 24 () :unsigned-fullword )
(define-syscall linux::stime 25 (:address) :signed-fullword )
(define-syscall linux::ptrace 26 (:unsigned-fullword
				  :unsigned-fullword
				  :address
				  :address)
		:signed-fullword)
(define-syscall linux::alarm 27 (:unsigned-fullword) :unsigned-fullword )

(define-syscall linux::pause 29 () :signed-fullword)
(define-syscall linux::utime 30 (:address :address) :signed-fullword )
(define-syscall linux::access 33 (:address :unsigned-fullword) :signed-fullword)
(define-syscall linux::nice 34 (:signed-fullword) :signed-fullword )
(define-syscall linux::sync 36 () :unsigned-fullword )
(define-syscall linux::kill 37 (:signed-fullword :unsigned-fullword) :signed-fullword )
(define-syscall linux::rename 38 (:address :address) :signed-fullword )
(define-syscall linux::mkdir 39 (:address :unsigned-fullword) :signed-fullword )
(define-syscall linux::rmdir 40 (:address) :signed-fullword )
(define-syscall linux::dup 41 (:unsigned-fullword) :signed-fullword )
(define-syscall linux::pipe 42 (:address) :signed-fullword )
(define-syscall linux::times 43 (:address) :unsigned-fullword )

(define-syscall linux::brk 45 (:address) :signed-fullword )
(define-syscall linux::setgid 46 (:unsigned-fullword) :signed-fullword)
(define-syscall linux::getgid 47 () :unsigned-fullword )
(define-syscall linux::signal 48 (:unsigned-fullword :address) :address )
(define-syscall linux::geteuid 49 () :unsigned-fullword )
(define-syscall linux::getegid 50 () :unsigned-fullword)
(define-syscall linux::acct 51 (:address) :signed-fullword )
(define-syscall linux::umount2 52 (:address) :signed-fullword )
(define-syscall linux::ioctl 54 (:unsigned-fullword :signed-fullword :address) :signed-fullword :min-args 2 )
(define-syscall linux::fcntl 55 (:unsigned-fullword :signed-fullword :signed-fullword) :signed-fullword :min-args 2 )
(define-syscall linux::setpgid 57 (:signed-fullword :signed-fullword) :signed-fullword )
(define-syscall linux::umask 60 (:unsigned-fullword) :unsigned-fullword )
(define-syscall linux::chroot 61 (:address) :signed-fullword )
(define-syscall linux::ustat 62 (:unsigned-fullword :address) :signed-fullword )
(define-syscall linux::dup2 63 (:unsigned-fullword :unsigned-fullword) :signed-fullword )
(define-syscall linux::getppid 64 () :unsigned-fullword)
(define-syscall linux::getpgrp 65 () :unsigned-fullword)
(define-syscall linux::setsid 66 () :signed-fullword)
(define-syscall linux::sigaction 67 (:unsigned-fullword :address :address) :signed-fullword )
(define-syscall linux::getrusage 77 (:signed-fullword :address) :signed-fullword)
(define-syscall linux::gettimeofday 78 (:address :address) :void)
(define-syscall linux::ftruncate 93 (:unsigned-fullword :unsigned-fullword)
		:signed-fullword)
(define-syscall linux::fchmod 94 (:unsigned-fullword :unsigned-fullword)
		:signed-fullword )
(define-syscall linux::socketcall 102 (:unsigned-fullword :address) :signed-fullword )
(define-syscall linux::stat 106 (:address :address) :signed-fullword)
(define-syscall linux::lstat 107 (:address :address) :signed-fullword)
(define-syscall linux::fstat 108 (:unsigned-fullword :address) :signed-fullword )

(define-syscall linux::fsync 118 (:unsigned-fullword) :signed-fullword )
(define-syscall linux::uname 122  (:address) :signed-fullword)
(define-syscall linux::fchdir 133 (:unsigned-fullword) :signed-fullword )
(define-syscall linux::_llseek 140 (:unsigned-fullword :unsigned-fullword :unsigned-fullword :address :unsigned-fullword) :signed-fullword )
(define-syscall linux::getcwd 182 (:address :unsigned-fullword) :signed-fullword )



#+notdefinedyet
(progn
(define-syscall linux::sgetmask 68 () )
(define-syscall linux::ssetmask 69 () )
(define-syscall linux::setreuid 70 () )
(define-syscall linux::setregid 71 () )
(define-syscall linux::sigsuspend 72 () )
(define-syscall linux::sigpending 73 () )
(define-syscall linux::sethostname 74 () )
(define-syscall linux::setrlimit 75 () )
(define-syscall linux::getrlimit 76 () )
(define-syscall linux::settimeofday 79 () )
(define-syscall linux::getgroups 80 () )
(define-syscall linux::setgroups 81 () )
(define-syscall linux::select 82 () )
(define-syscall linux::symlink 83 () )
(define-syscall linux::oldlstat 84 () )
(define-syscall linux::readlink 85 () )
(define-syscall linux::uselib 86 () )
(define-syscall linux::swapon 87 () )
(define-syscall linux::reboot 88 () )
(define-syscall linux::readdir 89 () )
(define-syscall linux::mmap 90 () )
(define-syscall linux::munmap 91 () )
(define-syscall linux::truncate 92 () )

(define-syscall linux::fchown 95 () )
(define-syscall linux::getpriority 96 () )
(define-syscall linux::setpriority 97 () )
(define-syscall linux::statfs 99 () )
(define-syscall linux::fstatfs 100 () )
(define-syscall linux::ioperm 101 () )
(define-syscall linux::syslog 103 () )
(define-syscall linux::setitimer 104 () )
(define-syscall linux::getitimer 105 () )
(define-syscall linux::olduname 109 () )
(define-syscall linux::iopl 110 () )
(define-syscall linux::vhangup 111 () )
(define-syscall linux::idle 112 () )
(define-syscall linux::vm86 113 () )
(define-syscall linux::wait4 114 () )
(define-syscall linux::swapoff 115 () )
(define-syscall linux::sysinfo 116 () )
(define-syscall linux::ipc 117 () )
(define-syscall linux::sigreturn 119 () )
(define-syscall linux::clone 120 () )
(define-syscall linux::setdomainname 121 () )
(define-syscall linux::modify_ldt 123 () )
(define-syscall linux::adjtimex 124 () )
(define-syscall linux::mprotect 125 () )
(define-syscall linux::sigprocmask 126 () )
(define-syscall linux::create_module	127 () )
(define-syscall linux::init_module	128 () )
(define-syscall linux::delete_module	129 () )
(define-syscall linux::get_kernel_syms	130 () )
(define-syscall linux::quotactl 131 () )
(define-syscall linux::getpgid 132 () )
(define-syscall linux::bdflush 134 () )
(define-syscall linux::sysfs 135 () )
(define-syscall linux::personality 136 () )
(define-syscall linux::setfsuid 138 () )
(define-syscall linux::setfsgid 139 () )
(define-syscall linux::getdents 141 () )
(define-syscall linux::_newselect 142 () )
(define-syscall linux::flock 143 () )
(define-syscall linux::msync 144 () )
(define-syscall linux::readv 145 () )
(define-syscall linux::writev 146 () )
(define-syscall linux::getsid 147 () )
(define-syscall linux::fdatasync 148 () )
(define-syscall linux::_sysctl 149 () )
(define-syscall linux::mlock 150 () )
(define-syscall linux::munlock 151 () )
(define-syscall linux::mlockall 152 () )
(define-syscall linux::munlockall 153 () )
(define-syscall linux::sched_setparam 154 () )
(define-syscall linux::sched_getparam 155 () )
(define-syscall linux::sched_setscheduler 156 () )
(define-syscall linux::sched_getscheduler 157 () )
(define-syscall linux::sched_yield 158 () )
(define-syscall linux::sched_get_priority_max 159 () )
(define-syscall linux::sched_get_priority_min 160 () )
(define-syscall linux::sched_rr_get_interval 161 () )
(define-syscall linux::nanosleep 162 () )
(define-syscall linux::mremap 163 () )
(define-syscall linux::setresuid 164 () )
(define-syscall linux::getresuid 165 () )
(define-syscall linux::query_module	166 () )
(define-syscall linux::poll 167 () )
(define-syscall linux::nfsservctl 168 () )
(define-syscall linux::setresgid 169 () )
(define-syscall linux::getresgid 170 () )
(define-syscall linux::prctl 171 () )
(define-syscall linux::rt_sigreturn 172 () )
(define-syscall linux::rt_sigaction 173 () )
(define-syscall linux::rt_sigprocmask 174 () )
(define-syscall linux::rt_sigpending 175 () )
(define-syscall linux::rt_sigtimedwait 176 () )
(define-syscall linux::rt_sigqueueinfo 177 () )
(define-syscall linux::rt_sigsuspend 178 () )
(define-syscall linux::pread 179 () )
(define-syscall linux::pwrite 180 () )
(define-syscall linux::chown 181 (:address) )

(define-syscall linux::capget 183 () )
(define-syscall linux::capset 184 () )
(define-syscall linux::sigaltstack 185 () )
(define-syscall linux::sendfile 186 () )
(define-syscall linux::getpmsg 187	 () )
(define-syscall linux::putpmsg 188	 () )
(define-syscall linux::vfork 189 () )

)